# Copyright 2016-2019 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os

from colcon_core.dependency_descriptor import DependencyDescriptor
from colcon_core.package_augmentation import PackageAugmentationExtensionPoint
from colcon_core.package_identification import IgnoreLocationException
from colcon_core.package_identification import logger
from colcon_core.package_identification \
    import PackageIdentificationExtensionPoint
from colcon_core.plugin_system import satisfies_version
from colcon_core.plugin_system import SkipExtensionException


# mapping paths to tuples containing the ROS package and its build type
_cached_packages = {}


class RosPackageIdentification(
    PackageIdentificationExtensionPoint, PackageAugmentationExtensionPoint
):
    """Identify ROS packages with `package.xml` files."""

    # the priority needs to be higher than the extensions identifying packages
    # using the build systems supported by ROS (CMake and Python)
    PRIORITY = 150

    def __init__(self):  # noqa: D107
        satisfies_version(
            PackageIdentificationExtensionPoint.EXTENSION_POINT_VERSION,
            '^1.0')
        satisfies_version(
            PackageAugmentationExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')
        # check if the necessary dependency to parse the manifest is available
        try:
            import catkin_pkg  # noqa: F401
        except ImportError:
            raise SkipExtensionException(
                "The Python module 'catkin_pkg' could not be imported, "
                'therefore ROS packages can not be identified')

    def identify(self, desc):  # noqa: D102
        # ignore packages which have been identified with a different type
        if desc.type is not None and desc.type != 'ros':
            return

        # skip paths with an ignore marker file
        if (desc.path / 'CATKIN_IGNORE').exists():
            raise IgnoreLocationException()
        if (desc.path / 'AMENT_IGNORE').exists():
            raise IgnoreLocationException()

        # parse package manifest and get build type
        pkg, build_type = get_package_with_build_type(str(desc.path))
        if not pkg or not build_type:
            # if it is not a wet ROS package check for a dry ROS package
            if (desc.path / 'manifest.xml').exists():
                # ignore location to avoid being identified as a CMake package
                raise IgnoreLocationException()
            return

        desc.type = f'ros.{build_type}'

        # use package name from manifest if not already set
        # e.g. from external configuration
        if desc.name is None:
            desc.name = pkg.name

    def augment_package(  # noqa: D102
        self, desc, *, additional_argument_names=None
    ):
        if not desc.type.startswith('ros.'):
            return

        pkg, _ = get_package_with_build_type(str(desc.path))
        if not pkg:
            return

        augment_package(desc, pkg)

    def augment_packages(  # noqa: D102
        self, descs, *, additional_argument_names=None
    ):
        super().augment_packages(
            descs, additional_argument_names=additional_argument_names)

        # get all parsed ROS package manifests
        global _cached_packages
        pkgs = {}
        for desc in descs:
            if str(desc.path) not in _cached_packages:
                continue
            pkg = _cached_packages[str(desc.path)][0]
            if pkg:
                pkgs[pkg] = desc

        add_group_dependencies(pkgs)


def augment_package(desc, pkg):
    """
    Augment a package descriptor based on data from a parsed package manifest.

    :param desc: The package descriptor
    :param pkg: The parsed package manifest
    """
    desc.metadata['version'] = pkg.version

    # get dependencies
    desc.dependencies['build'].update(
        _enumerate_dependency_descriptors(
            pkg.build_depends +
            pkg.buildtool_depends
        ))

    desc.dependencies['run'].update(
        _enumerate_dependency_descriptors(
            pkg.build_export_depends +
            pkg.buildtool_export_depends +
            pkg.exec_depends
        ))

    desc.dependencies['test'].update(
        _enumerate_dependency_descriptors(
            pkg.test_depends
        ))

    # get any maintainers which list an E-mail address
    maintainers = [str(m) for m in pkg.maintainers if m.email]
    if maintainers:
        desc.metadata.setdefault('maintainers', [])
        desc.metadata['maintainers'] += maintainers


def add_group_dependencies(pkgs):
    """
    Add appropriate dependencies based on group membership.

    :param dict pkgs: A mapping of parsed package manifests
      to associated package descriptors
    :
    """
    metadata = {
        'origin': 'ros',
    }

    # resolve group members and add them to the descriptor dependencies
    for pkg, desc in pkgs.items():
        for group_depend in pkg.group_depends:
            assert group_depend.evaluated_condition is not None
            if not group_depend.evaluated_condition:
                continue
            group_depend.extract_group_members(pkgs)
            for name in group_depend.members:
                desc.dependencies['build'].add(DependencyDescriptor(
                    name, metadata=metadata))
                desc.dependencies['run'].add(DependencyDescriptor(
                    name, metadata=metadata))


def get_package_with_build_type(path: str):
    """Get the ROS package and its build type for the given path."""
    global _cached_packages
    if path not in _cached_packages:
        pkg = _get_package(path)
        build_type = _get_build_type(pkg, path) if pkg else None
        _cached_packages[path] = (pkg, build_type)
    return _cached_packages[path]


def _get_package(path: str):
    """Get the ROS package for the given path."""
    from catkin_pkg.package import has_ros_schema_reference
    from catkin_pkg.package import InvalidPackage
    from catkin_pkg.package import package_exists_at
    from catkin_pkg.package import parse_package

    if not package_exists_at(path):
        return None

    try:
        pkg = parse_package(path)
    except (AssertionError, InvalidPackage) as e:  # noqa: F841
        if has_ros_schema_reference(path):
            logger.debug(
                f"Found ROS schema reference in package manifest in '{path}'")
            logger.warn(
                f"Failed to parse ROS package manifest in '{path}': {e}")
        else:
            logger.debug(
                'Failed to parse potential ROS package manifest in'
                f"'{path}': {e}")
        return None

    condition_context = {
        **os.environ,
        'DISABLE_GROUPS_WORKAROUND': '1',
    }
    pkg.evaluate_conditions(condition_context)
    return pkg


def _get_build_type(pkg, path):
    """Get the build type of the ROS package."""
    from catkin_pkg.package import InvalidPackage
    try:
        return pkg.get_build_type()
    except InvalidPackage:
        logger.warning(
            f"ROS package '{pkg.name}' in '{path}' has more than one "
            'build type')
        return None


def _enumerate_dependency_descriptors(dependencies):
    for dependency in dependencies:
        assert dependency.evaluated_condition is not None
        if not dependency.evaluated_condition:
            continue
        yield DependencyDescriptor(
            dependency.name, metadata=_create_metadata(dependency))


def _create_metadata(dependency):
    metadata = {
        'origin': 'ros',
    }
    attributes = (
        'version_lte',
        'version_lt',
        'version_gte',
        'version_gt',
        'version_eq',
    )
    for attr in attributes:
        if getattr(dependency, attr, None) is not None:
            metadata[attr] = getattr(dependency, attr)
    return metadata



# Copyright 2016-2018 Dirk Thomas
# Copyright 2024 Open Source Robotics Foundation, Inc.
# Licensed under the Apache License, Version 2.0

from pathlib import Path
from tempfile import TemporaryDirectory

from colcon_core.package_descriptor import PackageDescriptor
from colcon_core.package_identification import IgnoreLocationException
from colcon_ros.package_identification.ros import _cached_packages
from colcon_ros.package_identification.ros import RosPackageIdentification
import pytest


def test_identify():
    extension = RosPackageIdentification()
    augmentation_extension = extension

    with TemporaryDirectory(prefix='test_colcon_') as basepath:
        desc = PackageDescriptor(basepath)
        desc.type = 'other'
        assert extension.identify(desc) is None
        assert desc.name is None

        _cached_packages.clear()
        desc.type = None
        assert extension.identify(desc) is None
        assert desc.name is None
        assert desc.type is None

        _cached_packages.clear()
        basepath = Path(basepath)
        for marker in ('AMENT_IGNORE', 'CATKIN_IGNORE'):
            marker_path = basepath / marker
            marker_path.touch()
            with pytest.raises(IgnoreLocationException):
                extension.identify(desc)
            marker_path.unlink()

        _cached_packages.clear()
        (basepath / 'package.xml').write_text(
            '<package format="3">\n'
            '  <name>pkg-name</name>\n'
            '  <version>0.0.0</version>\n'
            '  <description>test package</description>\n'
            '  <maintainer email="foobar@example.com">Foo Bar</maintainer>\n'
            '  <license>Apache-2.0</license>\n'
            '</package>\n')
        assert extension.identify(desc) is None
        assert desc.name == 'pkg-name'
        assert desc.type == 'ros.catkin'
        assert not desc.dependencies
        assert not desc.metadata

        augmentation_extension.augment_packages([desc])
        assert desc.metadata['version'] == '0.0.0'
        assert set(desc.dependencies.keys()) == {'build', 'run', 'test'}
        assert not desc.dependencies['build']
        assert not desc.dependencies['run']
        assert not desc.dependencies['test']

        _cached_packages.clear()
        desc = PackageDescriptor(basepath)
        (basepath / 'package.xml').write_text(
            '<package format="3">\n'
            '  <name>other-name</name>\n'
            '  <version>0.0.0</version>\n'
            '  <description>test package</description>\n'
            '  <maintainer email="foobar@example.com">Foo Bar</maintainer>\n'
            '  <license>Apache-2.0</license>\n'
            '  <buildtool_depend>build</buildtool_depend>\n'
            '  <buildtool_export_depend version_gt="1.2.3">'
            'runA</buildtool_export_depend>\n'
            '  <test_depend>test</test_depend>\n'
            '  <exec_depend>runB</exec_depend>\n'
            '  <export>\n'
            '    <build_type>ament_cmake</build_type>\n'
            '  </export>\n'
            '</package>\n')
        assert extension.identify(desc) is None
        assert desc.name == 'other-name'
        assert desc.type == 'ros.ament_cmake'
        assert not desc.dependencies
        assert not desc.metadata

        augmentation_extension.augment_packages([desc])
        assert desc.metadata['version'] == '0.0.0'
        assert set(desc.dependencies.keys()) == {'build', 'run', 'test'}
        assert desc.dependencies['build'] == {'build'}
        assert desc.dependencies['run'] == {'runA', 'runB'}
        dep = next(x for x in desc.dependencies['run'] if x == 'runA')
        assert dep.metadata['version_gt'] == '1.2.3'
        assert desc.dependencies['test'] == {'test'}

        assert desc.metadata['maintainers'] == ['Foo Bar <foobar@example.com>']

        _cached_packages.clear()
        desc = PackageDescriptor(basepath)
        (basepath / 'manifest.xml').touch()
        assert extension.identify(desc) is None
        assert desc.type == 'ros.ament_cmake'

        _cached_packages.clear()
        desc = PackageDescriptor(basepath)
        (basepath / 'package.xml').unlink()
        with pytest.raises(IgnoreLocationException):
            extension.identify(desc)
        (basepath / 'manifest.xml').unlink()



# Copyright 2016-2019 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from colcon_core.package_augmentation import PackageAugmentationExtensionPoint
from colcon_core.package_identification.python import get_configuration
from colcon_core.package_identification.python import is_reading_cfg_sufficient
from colcon_core.plugin_system import satisfies_version
from colcon_python_setup_py.package_identification.python_setup_py \
    import get_setup_information


class RosAmentPythonPackageAugmentation(PackageAugmentationExtensionPoint):
    """Augment ament_python packages with information from setup files."""

    # the priority needs to be higher than the extensions augmenting packages
    # using the setup.py/setup.cfg files
    PRIORITY = 150

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            PackageAugmentationExtensionPoint.EXTENSION_POINT_VERSION,
            '^1.0')

    def augment_package(  # noqa: D102
        self, desc, *, additional_argument_names=None
    ):
        if desc.type != 'ros.ament_python':
            return

        if 'get_python_setup_options' in desc.metadata:
            return

        setup_py = desc.path / 'setup.py'
        if not setup_py.is_file():
            return

        setup_cfg = desc.path / 'setup.cfg'
        for _ in (1, ):
            # try to get information from setup.cfg file
            if setup_cfg.is_file():
                if is_reading_cfg_sufficient(setup_py):
                    config = get_configuration(setup_cfg)
                    name = config.get('metadata', {}).get('name')
                    if name:
                        options = config.get('options', {})

                        def getter(env):
                            nonlocal options
                            return options
                        break
        else:
            # use information from setup.py file

            def getter(env):  # noqa: F811
                nonlocal desc
                return get_setup_information(
                    desc.path / 'setup.py', env=env)

        desc.metadata['get_python_setup_options'] = getter



# Copyright 2016-2021 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from pathlib import Path

from colcon_core.shell import FindInstalledPackagesExtensionPoint


class AmentInstalledPackageFinder(FindInstalledPackagesExtensionPoint):
    """
    Find installed packages in an ament workspace.

    This is almost always /opt/ros/<rosdistro>
    """

    def find_installed_packages(self, install_base: Path):
        """Find installed packages in an ament workspace."""
        package_index = install_base.joinpath(
                'share', 'ament_index', 'resource_index', 'packages')

        if not package_index.is_dir():
            return None

        packages = {}
        for marker_path in package_index.iterdir():
            packages[marker_path.name] = install_base
        return packages



# Copyright 2022 Open Source Robotics Foundation, Inc.
# Licensed under the Apache License, Version 2.0

from pathlib import Path
from tempfile import TemporaryDirectory

from colcon_core.package_descriptor import PackageDescriptor
from colcon_ros.package_augmentation.ros_ament_index \
    import RosAmentIndexPackageAugmentation


def test_ament_index_augmentation():
    with TemporaryDirectory() as mock_prefix:
        mock_prefix = Path(mock_prefix)

        index_path = mock_prefix / 'share' / 'ament_index' / 'resource_index'
        dep_index_path = index_path / 'package_run_dependencies'
        dep_index_path.mkdir(parents=True)
        (dep_index_path / 'pkg_no_deps').write_text('')
        (dep_index_path / 'pkg_with_deps').write_text('dep1;dep2')

        extension = RosAmentIndexPackageAugmentation()

        # Non-existing
        desc = PackageDescriptor(mock_prefix)
        desc.name = 'pkg_not_in_index'
        desc.type = 'installed'
        extension.augment_packages((desc,))
        assert 'installed' == desc.type
        assert not desc.get_dependencies(categories=('run',))

        # Wrong type
        desc = PackageDescriptor(mock_prefix)
        desc.name = 'pkg_with_deps'
        desc.type = 'installed.other_type'
        extension.augment_packages((desc,))
        assert 'installed.other_type' == desc.type
        assert not desc.get_dependencies(categories=('run',))

        # Existing with deps
        desc = PackageDescriptor(mock_prefix)
        desc.name = 'pkg_with_deps'
        desc.type = 'installed'
        extension.augment_packages((desc,))
        assert 'installed.ros.ament' == desc.type
        assert {'dep1', 'dep2'} == desc.get_dependencies(categories=('run',))

        # Existing without deps
        desc = PackageDescriptor(mock_prefix)
        desc.name = 'pkg_no_deps'
        desc.type = 'installed'
        extension.augment_packages((desc,))
        assert 'installed.ros.ament' == desc.type
        assert not desc.get_dependencies(categories=('run',))



# Copyright 2022 Open Source Robotics Foundation, Inc.
# Licensed under the Apache License, Version 2.0

from colcon_core.dependency_descriptor import DependencyDescriptor
from colcon_core.package_augmentation \
    import PackageAugmentationExtensionPoint
from colcon_core.plugin_system import satisfies_version


class RosAmentIndexPackageAugmentation(PackageAugmentationExtensionPoint):
    """
    Augment installed packages with information from an ament index.

    Only packages of the `installed` type are considered.
    """

    # The priority needs to be higher than the extensions augmenting
    # 'installed' packages for the underlying package information sources
    # supported by ROS (colcon index and Python distributions)
    PRIORITY = 130

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            PackageAugmentationExtensionPoint.EXTENSION_POINT_VERSION,
            '^1.0')

    def augment_package(  # noqa: D102
        self, desc, *, additional_argument_names=None
    ):
        if desc.type != 'installed':
            return

        index_path = desc.path / 'share' / 'ament_index' / 'resource_index'
        marker_file = index_path / 'package_run_dependencies' / desc.name
        if not marker_file.is_file():
            return

        with marker_file.open() as f:
            raw_deps = [d for d in f.read().split(';') if d]
        desc.type = 'installed.ros.ament'
        desc.dependencies['run'].update(
            DependencyDescriptor(dep) for dep in raw_deps)
